import 'dart:math';

import 'package:bitcoin_flutter/bitcoin_flutter.dart';
import 'package:web3dart/crypto.dart';

import 'walletProxy.dart';

class BitcoinIn {
  String txHash;
  int vout;
  int value;

  BitcoinIn(this.txHash, this.vout, this.value);
}

class BitcoinTransaction {
  static Future<String> signBitcoinTransaction(
      String privateKey, Transaction transaction) async {
    if (transaction != null) {
      print(transaction.toHex());
      try {
        ECPair ecPair = ECPair.fromWIF(privateKey);
        TransactionBuilder transactionBuilder =
            TransactionBuilder.fromTransaction(transaction);
        transactionBuilder.inputs.forEach((element) {
          int index = transactionBuilder.inputs.indexOf(element);
          transactionBuilder.sign(vin: index, keyPair: ecPair);
        });
        String hex = transaction.toHex();
        return hex;
      } catch (e) {
        print('_signBitcoinTransaction error : $e');
      }
    } else {
      print('_signBitcoinTransaction error : transaction is null.');
    }
    return null;
  }

  // 6a = 106 OP_RETURN
  // 14 = 20
  // 6f6d6e69 = omni
  static const _omni = '6a146f6d6e69';
  static const int dirt = 546;
  static String _fillLengthWith(String s,
      {int length = 16, String char = '0'}) {
    if (s == null) return '';
    if (s.length >= length) return s.substring(0, length);
    return (char * length + s).substring(s.length);
  }

  static String _createOmniPayload(num value, int contract) {
    String type = contract.toRadixString(16);
    String amount = value.toInt().toRadixString(16);
    String payload = _omni + _fillLengthWith(type) + _fillLengthWith(amount);
    return payload;
  }

  static Future<Transaction> createBitcoinTransaction(
      String from, String to, num value, num fee,
      {String type = 'BTC', List<BitcoinIn> unspends}) async {
    if (WalletProxy.bitcoinType.keys.contains(type)) {
      CoinInfo coinInfo = WalletProxy.bitcoinType[type];
      try {
        final txb = new TransactionBuilder(network: coinInfo.network);
        int coinFee = (fee * pow(10, coinInfo.decimals)).toInt();
        int coinValue = (value * pow(10, coinInfo.decimals)).toInt();
        int needAmount = coinValue + coinFee;
        if (unspends == null) {
          //just for test
          unspends = _checkUnspends(from, needAmount);
        }
        int amount = 0;
        unspends.forEach((element) {
          amount += element.value;
          txb.addInput(element.txHash, element.vout);
        });
        int change = amount - coinValue - coinFee;
//        print(amount);
//        print(coinValue);
//        print(coinFee);
//        print(change);
        if (change < 0) {
          print('Amount is not enough.');
        } else {
          if (change >= dirt) {
            txb.addOutput(from, change);
          }
          txb.addOutput(to, coinValue);
          return txb.buildIncomplete();
        }
      } catch (e) {
        print('createBitcoinTransaction error : $e');
      }
    }
    return null;
  }

  static Future<Transaction> createOmniTransaction(
      String from, String to, num value, num fee,
      {String type = 'USDT', List<BitcoinIn> unspends}) async {
    if (WalletProxy.omniType.keys.contains(type)) {
      CoinInfo coinInfo = WalletProxy.omniType[type];
      try {
        final txb = new TransactionBuilder(network: coinInfo.network);
        int coinFee = (fee * pow(10, coinInfo.decimals)).toInt();
        int needAmount = coinFee + dirt;
        if (unspends == null) {
          //just for test
          unspends = _checkUnspends(from, needAmount);
        }
        int amount = 0;
        unspends.forEach((element) {
          amount += element.value;
          txb.addInput(element.txHash, element.vout);
        });
        int change = amount - dirt - coinFee;
//        print(amount);
//        print(dirt);
//        print(coinFee);
//        print(change);
        if (change < 0) {
          print('Amount is not enough.');
        } else {
          if (change >= dirt) {
            txb.addOutput(from, change);
          }
          txb.addOutput(to, dirt);

          int coinValue = (value * pow(10, coinInfo.decimals)).toInt();
          int contract = num.parse(coinInfo.address).toInt();
          String payload = _createOmniPayload(coinValue, contract);
          print(payload);

          txb.addOutput(hexToBytes(payload), 0);

          return txb.buildIncomplete();
        }
      } catch (e) {
        print('createOmniTransaction error : $e');
      }
    }
    return null;
  }

  static List<BitcoinIn> _checkUnspends(String from, int amount) {
    List<BitcoinIn> ins = List<BitcoinIn>()
      ..add(BitcoinIn(
          'a8a983ec1da3877947b4b15605952df0b948075d1b21ab62dfcd3bfbc73fc4ad',
          0,
          1111571));
    return ins;
  }
}
