import 'dart:math';
import 'dart:typed_data';

import 'package:walletpackage/src/rlp.dart';
import 'package:web3dart/crypto.dart';
import 'package:web3dart/web3dart.dart';

import 'walletProxy.dart';

class EthereumTransaction {
  static const int _ethChainId = 1;

  static Future<String> signEthereumTransaction(
      String privateKeyHex, Transaction transaction) async {
    if (transaction == null) {
      return null;
    } else {
      EthPrivateKey privateKey = EthPrivateKey.fromHex(privateKeyHex);
      final Uint8List unsignedTx =
          uint8ListFromList(encode(_encodeToRlp(transaction, null)));
//    print(web3dart_crypto.bytesToHex(unsignedTx));

      final signature =
          await privateKey.signToSignature(unsignedTx, chainId: _ethChainId);
      final Uint8List signedTx =
          uint8ListFromList(encode(_encodeToRlp(transaction, signature)));

      String signedTxhex = bytesToHex(signedTx);
//    print(signedTxhex);
      return signedTxhex;
    }
  }

  static List<dynamic> _encodeToRlp(
      Transaction transaction, MsgSignature signature) {
    final list = [
      transaction.nonce,
      transaction.gasPrice.getInWei,
      transaction.maxGas,
    ];

    if (transaction.to != null) {
      list.add(transaction.to.addressBytes);
    } else {
      list.add('');
    }

    list..add(transaction.value.getInWei)..add(transaction.data);

    if (signature != null) {
      list..add(signature.v)..add(signature.r)..add(signature.s);
    }
//    print(list);
    return list;
  }

  static Future<Transaction> createErc20Transaction(
      String from, String to, num value, int gasPrice, int nonce,
      {String type = 'EUSDT'}) async {
    if (WalletProxy.erc20Type.keys.contains(type)) {
      CoinInfo coinInfo = WalletProxy.erc20Type[type];

      String operationHex =
          '0xa9059cbb000000000000000000000000' + to.toLowerCase().substring(2);
      String valueHex =
          (value * pow(10, coinInfo.decimals)).toInt().toRadixString(16);
      String tokenHex =
          ('0000000000000000000000000000000000000000000000000000000000000000' +
                  valueHex)
              .substring(valueHex.length);
      String dataHex = operationHex + tokenHex;
//      print(dataHex);

      Transaction transaction = new Transaction(
          from: EthereumAddress.fromHex(from),
          to: EthereumAddress.fromHex(coinInfo.address),
          maxGas: coinInfo.gasLimit,
          gasPrice: EtherAmount.inWei(BigInt.from(gasPrice * 1000000000)),
          value: EtherAmount.zero(),
          data: hexToBytes(dataHex),
          nonce: nonce);

      return transaction;
    } else {
      return null;
    }
  }

  static Future<Transaction> createEthereumTransaction(String type, String from,
      String to, num value, int gasPrice, int nonce) async {
    if (WalletProxy.ethereumType.keys.contains(type)) {
      CoinInfo coinInfo = WalletProxy.erc20Type[type];

      Transaction transaction = new Transaction(
          from: EthereumAddress.fromHex(from),
          to: EthereumAddress.fromHex(coinInfo.address),
          maxGas: coinInfo.gasLimit,
          gasPrice: EtherAmount.inWei(BigInt.from(gasPrice * 1000000000)),
          value: EtherAmount.inWei(
              BigInt.from(value * pow(10, coinInfo.decimals))),
          data: null,
          nonce: nonce);

      return transaction;
    } else {
      return null;
    }
  }
}
